from typing import Any, Type, cast

from pydantic import BaseModel

from snowman.relation.table import (
    GenericColumnAccessor,
    GenericInsertColumnTypedDict,
    GenericOrderItemAccessor,
    GenericTable,
    GenericUpdateColumnTypedDict,
    Table,
)
from snowman.relation.table_like import TableLike


def full_table_name(table: Type[TableLike]) -> str:
    return f"{table.__database_name__}.{table.__schema_name__}.{table.__table_name__}"


def table_column_names(table: Type[TableLike]) -> list[str]:
    if issubclass(table, BaseModel):
        return [
            field.alias if field.alias else name
            for name, field in table.model_fields.items()
        ]

    else:
        raise ValueError(f"Table {table} is not a Pydantic model")


def table_columns_dict(
    table: Table[
        GenericTable,
        GenericColumnAccessor,
        GenericOrderItemAccessor,
        GenericInsertColumnTypedDict,
        GenericUpdateColumnTypedDict,
    ]
    | GenericInsertColumnTypedDict
    | GenericUpdateColumnTypedDict,
) -> dict[str, Any]:
    if isinstance(table, BaseModel):
        return table.model_dump()

    elif isinstance(table, dict):
        return cast(dict, table)

    else:
        raise ValueError(f"Table {table} is not a Pydantic model")



from typing import Any, Callable

from pydantic import Field


def DefaultFactory(default_factory: Callable[[], Any]) -> Any:
    """
    The wrapper for the combination of `Annotated` and `default_factory`.

    See: https://github.com/pydantic/pydantic/issues/9769
    """
    return Field(default_factory=default_factory)



from typing import Generic, TypeAlias

from snowflake.connector.cursor import SnowflakeCursor

from snowman.relation.table import GenericTable

try:
    import pandas as pd  # type: ignore[import]

    PandasDataFrame: TypeAlias = pd.DataFrame  # type: ignore
    USE_PANDAS = True

except ImportError:

    class DataFrame:
        pass

    PandasDataFrame: TypeAlias = DataFrame  # type: ignore
    USE_PANDAS = False


try:
    import turu.snowflake  # type: ignore[import]
    from turu.core.tag import Delete, Insert, Truncate, Update
    from turu.core.tag import Tag as TuruTag

    TuruSnowflakeCursor = turu.snowflake.Cursor  # type: ignore
    USE_TURU = True
    Tag = TuruTag  # type: ignore
    InsertTag = Insert  # type: ignore
    UpdateTag = Update  # type: ignore
    DeleteTag = Delete  # type: ignore
    TruncateTag = Truncate  # type: ignore


except ImportError:
    TuruSnowflakeCursor = SnowflakeCursor  # type: ignore
    USE_TURU = False

    class Tag:  # type: ignore
        pass

    class InsertTag(Tag, Generic[GenericTable]):  # type: ignore
        pass

    class UpdateTag(Tag, Generic[GenericTable]):  # type: ignore
        pass

    class DeleteTag(Tag, Generic[GenericTable]):  # type: ignore
        pass

    class TruncateTag(Tag, Generic[GenericTable]):  # type: ignore
        pass



from typing import Any, Generic, Type, cast

from pydantic.fields import FieldInfo
from typing_extensions import override

from snowman._generic import PyType
from snowman.query.column import Column, GenericColumnName
from snowman.query.to_sql import OperationWithParams, ToSql
from snowman.relation.table import (
    GenericColumnAccessor,
    GenericInsertColumnTypedDict,
    GenericOrderItemAccessor,
    GenericUpdateColumnTypedDict,
    Table,
)
from snowman.relation.table_like import GenericTableLike
from snowman.relation.view import View


class OrderItem(Generic[GenericTableLike, GenericColumnName, PyType], ToSql):
    pass


class ColumnOrderItem(
    OrderItem[GenericTableLike, GenericColumnName, PyType],
):
    def __init__(self, column: Column[GenericTableLike, GenericColumnName, PyType]):
        self._column = column

    @property
    def asc(self) -> "AscOrderItem[GenericTableLike, GenericColumnName, PyType]":
        return AscOrderItem(self)

    @property
    def desc(self) -> "DescOrderItem[GenericTableLike, GenericColumnName, PyType]":
        return DescOrderItem(self)

    @property
    def nulls(self) -> "NullsOrder[GenericTableLike, GenericColumnName, PyType]":
        return NullsOrder(self)

    @override
    def to_sql(self) -> OperationWithParams:
        return OperationWithParams(
            operation=str(self._column),
            params=(),
        )

    def __str__(self) -> str:
        return str(self._column)

    def __repr__(self) -> str:
        return repr(self._column)


class AscOrderItem(OrderItem[GenericTableLike, GenericColumnName, PyType]):
    def __init__(
        self, order_item: OrderItem[GenericTableLike, GenericColumnName, PyType]
    ):
        self._order_item = order_item

    @property
    def nulls(self) -> "NullsOrder[GenericTableLike, GenericColumnName, PyType]":
        return NullsOrder(self)

    @override
    def to_sql(self) -> OperationWithParams:
        operation, params = self._order_item.to_sql()
        return OperationWithParams(
            operation=f"{operation} ASC",
            params=params,
        )


class DescOrderItem(OrderItem[GenericTableLike, GenericColumnName, PyType]):
    def __init__(
        self, order_item: OrderItem[GenericTableLike, GenericColumnName, PyType]
    ):
        self._order_item = order_item

    @property
    def nulls(self) -> "NullsOrder[GenericTableLike, GenericColumnName, PyType]":
        return NullsOrder(self)

    @override
    def to_sql(self) -> OperationWithParams:
        operation, params = self._order_item.to_sql()
        return OperationWithParams(
            operation=f"{operation} DESC",
            params=params,
        )


class NullsOrder(Generic[GenericTableLike, GenericColumnName, PyType]):
    def __init__(self, item: OrderItem):
        self._item = item

    @property
    def first(
        self,
    ) -> "NullsFirstOrderItem[GenericTableLike, GenericColumnName, PyType]":
        return NullsFirstOrderItem(self._item)

    @property
    def last(self) -> "NullsLastOrderItem[GenericTableLike, GenericColumnName, PyType]":
        return NullsLastOrderItem(self._item)


class NullsFirstOrderItem(OrderItem[GenericTableLike, GenericColumnName, PyType]):
    def __init__(self, item: OrderItem):
        self._item = item

    @override
    def to_sql(self) -> OperationWithParams:
        operation, params = self._item.to_sql()
        return OperationWithParams(
            operation=f"{operation} NULLS FIRST",
            params=params,
        )


class NullsLastOrderItem(OrderItem[GenericTableLike, GenericColumnName, PyType]):
    def __init__(self, item: OrderItem):
        self._item = item

    @override
    def to_sql(self) -> OperationWithParams:
        operation, params = self._item.to_sql()

        return OperationWithParams(
            operation=f"{operation} NULLS LAST",
            params=params,
        )


class _InternalOrderItemAccessor:
    """
    An internal implementation class for accessing column information.

    On the editor, this class behaves as GenericColumnAccessor,
    but this class is called when accessing properties.
    """

    def __init__(
        self,
        table: Type[
            Table[
                GenericTableLike,
                GenericColumnAccessor,
                GenericOrderItemAccessor,
                GenericInsertColumnTypedDict,
                GenericUpdateColumnTypedDict,
            ]
            | View[GenericColumnAccessor, GenericOrderItemAccessor]
        ],
    ):
        self._table = table

    def __getattr__(self, key: str) -> OrderItem[Any, Any, Any]:
        field: FieldInfo = self._table.model_fields[key]
        return ColumnOrderItem(
            Column(
                cast(type, field.annotation),
                database_name=self._table.__database_name__,
                schema_name=self._table.__schema_name__,
                table_name=self._table.__table_name__,
                column_name=field.alias if field.alias else key,
            )
        )



from typing import TYPE_CHECKING

from typing_extensions import override

from snowman._generic import PyType
from snowman.query.builder.condition.condition import Condition
from snowman.query.to_sql import OperationWithParams

if TYPE_CHECKING:
    from snowman.query.column import Column, GenericColumnName
    from snowman.relation.table_like import GenericTableLike


class IsNotCondition(Condition):
    def __init__(self, base: "Column[GenericTableLike, GenericColumnName,PyType]"):
        self._base = base

    @override
    def to_sql(self) -> OperationWithParams:
        return OperationWithParams(operation=f"{self._base} IS NOT NULL", params=())



from typing import TYPE_CHECKING, Sequence

from snowman._generic import PyType
from snowman.query.builder.condition.condition import Condition
from snowman.query.to_sql import OperationWithParams

if TYPE_CHECKING:
    from snowman.query.column import Column, GenericColumnName
    from snowman.relation.table_like import GenericTableLike


class NotInCondition(Condition):
    def __init__(
        self,
        base: "Column[GenericTableLike, GenericColumnName, PyType]",
        values: Sequence[PyType],
    ):
        self._base = base
        self._values = values

    def to_sql(self) -> OperationWithParams:
        return OperationWithParams(
            operation=f"{self._base} NOT IN (%s)", params=(self._values,)
        )



from typing import Callable, ClassVar, Generic, LiteralString, Type, TypeVar

import pydantic

from snowman.relation.table import GenericColumnAccessor, GenericOrderItemAccessor


class View(
    pydantic.BaseModel, Generic[GenericColumnAccessor, GenericOrderItemAccessor]
):
    """
    Base class for view definition.
    """

    # NOTE: Meta information to specify the table name accurately.
    #       It is used when creating a query.
    __database_name__: ClassVar[str]
    __schema_name__: ClassVar[str]
    __table_name__: ClassVar[str]

    # NOTE: This field exists only for type definition and is not accessed at runtime.
    __access_columns__: Type[GenericColumnAccessor] | None = None

    model_config = pydantic.ConfigDict(populate_by_name=True)


GenericView = TypeVar("GenericView", bound=View)


def view(
    database_name: LiteralString,
    schema_name: LiteralString,
    table_name: LiteralString,
    /,
) -> "Callable[[Type[GenericView]], Type[GenericView]]":
    def decorate(cls: "Type[GenericView]") -> "Type[GenericView]":
        cls.__database_name__ = database_name
        cls.__schema_name__ = schema_name
        cls.__table_name__ = table_name

        return cls

    return decorate



from typing import (
    TYPE_CHECKING,
    Any,
    Generic,
    LiteralString,
    Sequence,
    Type,
    TypeVar,
    cast,
    overload,
)

from pydantic.fields import FieldInfo

from snowman._generic import PyType
from snowman.query.builder.condition.eq_condition import EqCondition
from snowman.query.builder.condition.ge_condition import GeCondition
from snowman.query.builder.condition.gt_condition import GtCondition
from snowman.query.builder.condition.in_condition import InCondition
from snowman.query.builder.condition.is_condition import IsCondition
from snowman.query.builder.condition.is_not_condition import IsNotCondition
from snowman.query.builder.condition.le_condition import LeCondition
from snowman.query.builder.condition.lt_condition import LtCondition
from snowman.query.builder.condition.ne_condition import NeCondition
from snowman.query.builder.condition.not_in_condition import NotInCondition
from snowman.relation.table import (
    GenericColumnAccessor,
    GenericInsertColumnTypedDict,
    GenericOrderItemAccessor,
    GenericUpdateColumnTypedDict,
    Table,
)
from snowman.relation.table_like import GenericTableLike
from snowman.relation.view import View

if TYPE_CHECKING:
    from snowman.typing import (
        TypeMissMatch,
        U,
        UseIsInsteadOfEq,
        UseIsNotInsteadOfNe,
    )

GenericColumnName = TypeVar("GenericColumnName", bound=LiteralString)


class Column(Generic[GenericTableLike, GenericColumnName, PyType]):
    def __init__(
        self,
        data_type: Type[PyType],
        /,
        *,
        database_name: str,
        schema_name: str,
        table_name: str,
        column_name: str,
    ):
        self._data_type = data_type
        self._database_name = database_name
        self._schema_name = schema_name
        self._table_name = table_name
        self._column_name = column_name

    @property
    def is_(self) -> "ColumnIs[GenericTableLike, GenericColumnName, PyType]":
        return ColumnIs(self)

    def in_(self, values: Sequence[PyType], /) -> InCondition:
        return InCondition(self, values)

    @property
    def not_(self) -> "ColumnNot[GenericTableLike, GenericColumnName, PyType]":
        return ColumnNot(self)

    @overload  # type: ignore
    def __eq__(self, value: bool) -> "UseIsInsteadOfEq": ...  # type: ignore
    @overload
    def __eq__(self, value: None) -> "UseIsInsteadOfEq": ...  # type: ignore
    @overload
    def __eq__(self, value: PyType) -> EqCondition: ...  # type: ignore
    @overload
    def __eq__(self, value: "U") -> "TypeMissMatch[PyType, U]": ...
    def __eq__(self, value: PyType) -> EqCondition:  # type: ignore
        return EqCondition(self, value)

    @overload  # type: ignore
    def __ne__(self, value: bool) -> "UseIsNotInsteadOfNe": ...  # type: ignore
    @overload
    def __ne__(self, value: None) -> "UseIsNotInsteadOfNe": ...  # type: ignore
    @overload
    def __ne__(self, value: PyType) -> NeCondition: ...  # type: ignore
    @overload
    def __ne__(self, value: "U") -> "TypeMissMatch[PyType, U]": ...
    def __ne__(self, value: PyType) -> NeCondition:  # type: ignore
        return NeCondition(self, value)

    @overload
    def __le__(self, value: PyType) -> LeCondition: ...  # type: ignore
    @overload
    def __le__(self, value: "U") -> "TypeMissMatch[PyType, U]": ...
    def __le__(self, value: PyType) -> LeCondition:  # type: ignore
        return LeCondition(self, value)

    @overload
    def __ge__(self, value: PyType) -> GeCondition: ...  # type: ignore
    @overload
    def __ge__(self, value: "U") -> "TypeMissMatch[PyType, U]": ...
    def __ge__(self, value: PyType) -> GeCondition:  # type: ignore
        return GeCondition(self, value)

    @overload
    def __lt__(self, value: PyType) -> LtCondition: ...  # type: ignore
    @overload
    def __lt__(self, value: "U") -> "TypeMissMatch[PyType, U]": ...
    def __lt__(self, value: PyType) -> LtCondition:  # type: ignore
        return LtCondition(self, value)

    @overload
    def __gt__(self, value: PyType) -> GtCondition: ...  # type: ignore
    @overload
    def __gt__(self, value: "U") -> "TypeMissMatch[PyType, U]": ...
    def __gt__(self, value: PyType) -> GtCondition:  # type: ignore
        return GtCondition(self, value)

    def __str__(self) -> str:
        return self._column_name

    def __repr__(self) -> str:
        return f"{self._database_name}.{self._schema_name}.{self._table_name}.{self._column_name}"


class _InternalColumnAccessor:
    """
    An internal implementation class for accessing column information.

    On the editor, this class behaves as GenericColumnAccessor,
    but this class is called when accessing properties.
    """

    def __init__(
        self,
        table: Type[
            Table[
                GenericTableLike,
                GenericColumnAccessor,
                GenericOrderItemAccessor,
                GenericInsertColumnTypedDict,
                GenericUpdateColumnTypedDict,
            ]
            | View[GenericColumnAccessor, GenericOrderItemAccessor]
        ],
    ):
        self._table = table

    def __getattr__(self, key: str) -> Column[Any, Any, Any]:
        field: FieldInfo = self._table.model_fields[key]
        return Column(
            cast(type, field.annotation),
            database_name=self._table.__database_name__,
            schema_name=self._table.__schema_name__,
            table_name=self._table.__table_name__,
            column_name=field.alias if field.alias else key,
        )


class ColumnIs(Generic[GenericTableLike, GenericColumnName, PyType]):
    """
    A class for using the IS operator

    #### Usage:
        ```python
        User.id.is_.true
        User.id.is_.false
        User.id.is_.null
        User.id.is_(True)
        ```
    """

    def __init__(self, column: Column[GenericTableLike, GenericColumnName, PyType]):
        self._column = column

    @property
    def not_(self) -> "ColumnIsNot":
        return ColumnIsNot(self._column)

    @property
    def null(self) -> IsCondition:
        return IsCondition(self._column)


class ColumnIsNot(Generic[PyType]):
    """
    A class for using the IS NOT operator

    #### Usage:
        ```python
        User.id.is_.not_.true
        User.id.is_.not_.false
        User.id.is_.not_.null
        User.id.is_.not_(True)
        ```
    """

    def __init__(self, column: Column[GenericTableLike, GenericColumnName, PyType]):
        self._column = column

    @property
    def null(self) -> "IsNotCondition":
        return IsNotCondition(self._column)


class ColumnNot(Generic[GenericTableLike, GenericColumnName, PyType]):
    """
    A class for using the NOT operator

    #### Usage:
        ```python
        User.id.not_.in_([1, 2, 3])
        ```
    """

    def __init__(self, column: Column[GenericTableLike, GenericColumnName, PyType]):
        self._column = column

    def in_(self, values: Sequence[PyType], /) -> NotInCondition:
        return NotInCondition(self._column, values)



import re

SPACE_PATTERN = re.compile(r"\s+")


def minify(query: str) -> str:
    """
    Minify the given query.

    #### Usage:
        ```python
        minify("SELECT * FROM database.schema.users")
        ```
    """
    return re.sub(SPACE_PATTERN, " ", query.replace("\n", "")).strip()



from abc import abstractmethod


class snowmanException(Exception):
    @property
    @abstractmethod
    def message(self) -> str: ...

    def __str__(self) -> str:
        return self.message


class snowmanError(snowmanException):
    pass


class snowmanNotDataFrameAvailableError(snowmanError):
    @property
    def message(self) -> str:
        return "This method is not available when using DataFrame."



from snowflake.connector import SnowflakeConnection
from snowflake.connector.cursor import SnowflakeCursor

from snowman._features import TuruSnowflakeCursor

Cursor = SnowflakeCursor | TuruSnowflakeCursor


def _get_snowfalke_cursor(cursor: Cursor) -> SnowflakeCursor:
    if raw_cursor := getattr(cursor, "_raw_cursor", None):
        return raw_cursor

    else:
        return cursor  # type: ignore


def _get_snowflake_connection(
    cursor: Cursor,
) -> SnowflakeConnection:
    cursor = _get_snowfalke_cursor(cursor)
    return getattr(cursor, "connection")



from typing import Generic, Type

from snowman.query.expression import order
from snowman.relation.table import (
    GenericColumnAccessor,
    GenericInsertColumnTypedDict,
    GenericOrderItemAccessor,
    GenericUpdateColumnTypedDict,
    Table,
)
from snowman.relation.table_like import GenericTableLike
from snowman.relation.view import View


class OrderByContext(
    Generic[
        GenericTableLike,
        GenericColumnAccessor,
        GenericOrderItemAccessor,
        GenericInsertColumnTypedDict,
        GenericUpdateColumnTypedDict,
    ]
):
    def __init__(
        self,
        table: Type[
            Table[
                GenericTableLike,
                GenericColumnAccessor,
                GenericOrderItemAccessor,
                GenericInsertColumnTypedDict,
                GenericUpdateColumnTypedDict,
            ]
            | View[
                GenericColumnAccessor,
                GenericOrderItemAccessor,
            ]
        ],
    ) -> None:
        self._table = table

    def __call__(
        self,
        table: Type[
            Table[
                GenericTableLike,
                GenericColumnAccessor,
                GenericOrderItemAccessor,
                GenericInsertColumnTypedDict,
                GenericUpdateColumnTypedDict,
            ]
            | View[
                GenericColumnAccessor,
                GenericOrderItemAccessor,
            ]
        ],
    ) -> GenericOrderItemAccessor:
        return order(table)

    @property
    def self(self) -> GenericOrderItemAccessor:
        return order(self._table)



from typing import (
    Callable,
    ClassVar,
    Generic,
    LiteralString,
    Type,
    TypedDict,
    TypeVar,
)

import pydantic


class AccessColumnTypedDict(TypedDict):
    pass


class InsertColumnTypedDict(TypedDict):
    pass


class UpdateColumnTypedDict(TypedDict, total=False):
    pass


_GenericTable = TypeVar("_GenericTable")

GenericColumnAccessor = TypeVar(
    "GenericColumnAccessor",
)

GenericOrderItemAccessor = TypeVar(
    "GenericOrderItemAccessor",
)

GenericInsertColumnTypedDict = TypeVar(
    "GenericInsertColumnTypedDict",
    bound=InsertColumnTypedDict,
)

GenericUpdateColumnTypedDict = TypeVar(
    "GenericUpdateColumnTypedDict",
    bound=UpdateColumnTypedDict,
)


class Table(
    pydantic.BaseModel,
    Generic[
        _GenericTable,
        GenericColumnAccessor,
        GenericOrderItemAccessor,
        GenericInsertColumnTypedDict,
        GenericUpdateColumnTypedDict,
    ],
):
    """
    Base class for table definition.
    """

    # NOTE: Meta information to specify the table name accurately.
    #       It is used when creating a query.
    __database_name__: ClassVar[str]
    __schema_name__: ClassVar[str]
    __table_name__: ClassVar[str]

    # NOTE: This field exists only for type definition and is not accessed at runtime.
    __access_columns__: Type[GenericColumnAccessor] | None = None
    __insert_columns__: Type[GenericInsertColumnTypedDict] | None = None
    __update_columns__: Type[GenericUpdateColumnTypedDict] | None = None

    model_config = pydantic.ConfigDict(populate_by_name=True)


GenericTable = TypeVar("GenericTable", bound=Table)


def table(
    database_name: LiteralString,
    schema_name: LiteralString,
    table_name: LiteralString,
    /,
) -> "Callable[[Type[GenericTable]], Type[GenericTable]]":
    def decorate(cls: "Type[GenericTable]") -> "Type[GenericTable]":
        cls.__database_name__ = database_name
        cls.__schema_name__ = schema_name
        cls.__table_name__ = table_name

        return cls

    return decorate



from typing import TYPE_CHECKING

from snowman.query.to_sql import ToSql

if TYPE_CHECKING:
    from snowman.query.builder.condition.and_condition import AndCondition
    from snowman.query.builder.condition.or_condition import OrCondition


class Condition(ToSql):
    def and_(self, other: "Condition") -> "AndCondition":
        from snowman.query.builder.condition.and_condition import AndCondition

        return AndCondition(self, other)

    def or_(self, other: "Condition") -> "OrCondition":
        from snowman.query.builder.condition.or_condition import OrCondition

        return OrCondition(self, other)



"""
# ⛄ [Snowman](https://github.com/ya7010/snowman-py) ⛄

Python model and query builder for [Snowflake](https://www.snowflake.com/).
"""

import importlib
import importlib.metadata

from . import datatype as datatype
from . import pydantic as pydantic
from . import query as query
from .query.column import Column as Column
from .query.order_item import ColumnOrderItem as ColumnOrderItem
from .relation.table import Table as Table
from .relation.table import table as table
from .relation.view import View as View
from .relation.view import view as view

__version__ = importlib.metadata.version("snowman-py")



"""
Snowman provides simple query builder for `insert` / `update` / `delete` / `truncate`.
"""

from typing import Type

from snowman.query.builder._delete_query_builder import DeleteQueryBuilder
from snowman.query.builder._insert_query_builder import InsertQueryBuilder
from snowman.query.builder._select_query_builder import SelectQueryBuilder
from snowman.query.builder._truncate_query_builder import TruncateQueryBuilder
from snowman.query.builder._update_query_builder import UpdateStatement
from snowman.relation.table import (
    GenericColumnAccessor,
    GenericInsertColumnTypedDict,
    GenericOrderItemAccessor,
    GenericTable,
    GenericUpdateColumnTypedDict,
    Table,
)

insert = InsertQueryBuilder()
"""
Insert query builder.
"""


def update(
    table: Type[
        Table[
            GenericTable,
            GenericColumnAccessor,
            GenericOrderItemAccessor,
            GenericInsertColumnTypedDict,
            GenericUpdateColumnTypedDict,
        ]
    ],
):
    """
    Update query builder.
    """
    return UpdateStatement(
        table,
        _columns_type=table.__update_columns__,  # type: ignore
    )


delete = DeleteQueryBuilder()
"""
Delete query builder.
"""

truncate = TruncateQueryBuilder()
"""
Truncate query builder.
"""


select = SelectQueryBuilder
"""
Select query builder.

🚧 🚧 🚧 **This is a draft.** This interface is subject to disruptive changes. 🚧 🚧 🚧
"""



from typing import Generic, Type

from typing_extensions import override

from snowman._cursor import Cursor
from snowman._features import TruncateTag
from snowman.relation import full_table_name
from snowman.relation.table import GenericTable

from ._builder import QueryBuilder, QueryWithParams, execute_with_tag


class TruncateQueryBuilder:
    @property
    def table(self) -> "TruncateTableStatement":
        return TruncateTableStatement()

    def __call__(
        self, table: Type[GenericTable], /
    ) -> "TruncateTableQueryBuilder[GenericTable]":
        return TruncateTableQueryBuilder(table, if_exists=False)

    @property
    def if_(self) -> "TruncateTableIfStatement":
        return TruncateTableIfStatement()


class TruncateTableStatement:
    def __call__(
        self, table: Type[GenericTable], /
    ) -> "TruncateTableQueryBuilder[GenericTable]":
        return TruncateTableQueryBuilder(table, if_exists=False)

    @property
    def if_(self) -> "TruncateTableIfStatement":
        return TruncateTableIfStatement()


class TruncateTableIfStatement:
    def exists(
        self, table: Type[GenericTable], /
    ) -> "TruncateTableQueryBuilder[GenericTable]":
        return TruncateTableQueryBuilder(table, if_exists=True)


class TruncateTableQueryBuilder(QueryBuilder[None], Generic[GenericTable]):
    def __init__(self, table: Type[GenericTable], /, *, if_exists: bool):
        self._table = table
        self._if_exists = if_exists

    @override
    def build(self) -> QueryWithParams:
        if_exists = " IF EXISTS" if self._if_exists else ""
        query = f"TRUNCATE TABLE{if_exists} {full_table_name(self._table)}"

        return QueryWithParams(query, ())

    @override
    def execute(self, cursor: Cursor) -> None:
        query, params = self.build()
        execute_with_tag(
            TruncateTag[self._table],  # type: ignore
            cursor,
            query,
            params,
        )



from typing import TYPE_CHECKING

from snowman.query.builder.condition.condition import Condition
from snowman.query.to_sql import OperationWithParams

if TYPE_CHECKING:
    pass


class GroupCondition(Condition):
    def __init__(self, condition: Condition, /):
        self._condition = condition

    def to_sql(self) -> OperationWithParams:
        sql = self._condition.to_sql()
        return OperationWithParams(operation=f"({sql.operation})", params=sql.params)



from typing_extensions import override

from snowman.query.builder.condition.condition import Condition
from snowman.query.to_sql import OperationWithParams


class OrCondition(Condition):
    def __init__(self, left: Condition, right: Condition):
        self._left = left
        self._right = right

    @override
    def to_sql(self) -> OperationWithParams:
        left = self._left.to_sql()
        right = self._right.to_sql()
        return OperationWithParams(
            f"{left.operation} OR {right.operation}",
            left.params + right.params,
        )



import datetime
import decimal
from typing import Any

NUMBER = float
DECIMAL = decimal.Decimal
NUMERIC = float
INT = int
INTEGER = int
BIGINT = int
SMALLINT = int
TINYINT = int
BYTEINT = int
FLOAT = float
FLOAT4 = float
FLOAT8 = float
DOUBLE = float
REAL = float
VARCHAR = str
CHAR = str
CHARACTER = str
STRING = str
TEXT = str
BINARY = bytes
VARBINARY = bytes
BOOLEAN = bool
DATE = datetime.date
DATETIME = datetime.datetime
TIME = datetime.time
TIMESTAMP = datetime.datetime
TIMESTAMP_LTZ = datetime.datetime
TIMESTAMP_NTZ = datetime.datetime
TIMESTAMP_TZ = datetime.datetime
VARIANT = Any
OBJECT = dict[str, Any]
ARRAY = list[Any]
GEOGRAPHY = Any
GEOMETRY = Any



import itertools
from typing import Any, Callable, Generic, Sequence, Type, cast, overload

from pydantic import BaseModel
from typing_extensions import override

from snowman._cursor import Cursor
from snowman._features import UpdateTag
from snowman.query.builder.condition.condition import Condition
from snowman.query.builder.context.where_context import WhereContext
from snowman.relation import full_table_name
from snowman.relation.table import (
    GenericColumnAccessor,
    GenericInsertColumnTypedDict,
    GenericOrderItemAccessor,
    GenericTable,
    GenericUpdateColumnTypedDict,
    Table,
)

from ._builder import QueryBuilder, QueryWithParams, execute_with_tag


class UpdateStatement(
    Generic[
        GenericTable,
        GenericColumnAccessor,
        GenericOrderItemAccessor,
        GenericInsertColumnTypedDict,
        GenericUpdateColumnTypedDict,
    ]
):
    def __init__(
        self,
        table: Type[
            Table[
                GenericTable,
                GenericColumnAccessor,
                GenericOrderItemAccessor,
                GenericInsertColumnTypedDict,
                GenericUpdateColumnTypedDict,
            ]
        ],
        /,
        *,
        _columns_type: Type[GenericUpdateColumnTypedDict] | None = None,
    ) -> None:
        self._table = table

    def set(
        self,
        fields: Table[
            GenericTable,
            GenericColumnAccessor,
            GenericOrderItemAccessor,
            GenericInsertColumnTypedDict,
            GenericUpdateColumnTypedDict,
        ]
        | GenericUpdateColumnTypedDict,
    ) -> "UpdateSetQueryBuilder[GenericTable, GenericColumnAccessor, GenericOrderItemAccessor, GenericInsertColumnTypedDict, GenericUpdateColumnTypedDict]":
        return UpdateSetQueryBuilder(self._table, fields)


class UpdateSetQueryBuilder(
    Generic[
        GenericTable,
        GenericColumnAccessor,
        GenericOrderItemAccessor,
        GenericInsertColumnTypedDict,
        GenericUpdateColumnTypedDict,
    ]
):
    def __init__(
        self,
        table: type[
            Table[
                GenericTable,
                GenericColumnAccessor,
                GenericOrderItemAccessor,
                GenericInsertColumnTypedDict,
                GenericUpdateColumnTypedDict,
            ]
        ],
        columns: Table[
            GenericTable,
            GenericColumnAccessor,
            GenericOrderItemAccessor,
            GenericInsertColumnTypedDict,
            GenericUpdateColumnTypedDict,
        ]
        | GenericUpdateColumnTypedDict,
    ):
        self._table = table
        self._columns = columns

    @overload
    def where(
        self,
        condition: Callable[[WhereContext[GenericColumnAccessor]], Condition],
        /,
    ) -> "UpdateSetWhereQueryBuidler[GenericTable, GenericColumnAccessor, GenericOrderItemAccessor, GenericInsertColumnTypedDict, GenericUpdateColumnTypedDict]": ...

    @overload
    def where(
        self,
        condition: Condition,
        /,
    ) -> "UpdateSetWhereQueryBuidler[GenericTable, GenericColumnAccessor, GenericOrderItemAccessor, GenericInsertColumnTypedDict, GenericUpdateColumnTypedDict]": ...

    @overload
    def where(
        self,
        condition: str,
        params: Sequence[Any] | None = None,
        /,
    ) -> "UpdateSetWhereQueryBuidler[GenericTable, GenericColumnAccessor, GenericOrderItemAccessor, GenericInsertColumnTypedDict, GenericUpdateColumnTypedDict]": ...

    def where(
        self,
        condition: Callable[[WhereContext[GenericColumnAccessor]], Condition]
        | Condition
        | str,
        params: Sequence[Any] | None = None,
        /,
    ) -> "UpdateSetWhereQueryBuidler[GenericTable, GenericColumnAccessor, GenericOrderItemAccessor, GenericInsertColumnTypedDict, GenericUpdateColumnTypedDict]":
        """
        Specify the condition of the where clause.

        Query parameters only support positional placeholders, so specify them with `%s`.

        e.g)
            - `.where(lambda c: (c(User).id == 1).and_(c(User).name == "Alice"))`
            - `.where((column(User).id == 1).and_(column(User).name == "Alice"))`
            - `.where("id = %s AND name = %s", [1, "Alice"])`
        """
        if callable(condition):
            condition, params = condition(WhereContext(self._table)).to_sql()

        elif isinstance(condition, Condition):
            condition, params = condition.to_sql()

        return UpdateSetWhereQueryBuidler(
            self._table,
            self._columns,
            where_condition=condition,
            where_params=params or (),
        )


class UpdateSetWhereQueryBuidler(
    QueryBuilder[None],
    Generic[
        GenericTable,
        GenericColumnAccessor,
        GenericOrderItemAccessor,
        GenericInsertColumnTypedDict,
        GenericUpdateColumnTypedDict,
    ],
):
    def __init__(
        self,
        table: type[
            Table[
                GenericTable,
                GenericColumnAccessor,
                GenericOrderItemAccessor,
                GenericInsertColumnTypedDict,
                GenericUpdateColumnTypedDict,
            ]
        ],
        columns: Table[
            GenericTable,
            GenericColumnAccessor,
            GenericOrderItemAccessor,
            GenericInsertColumnTypedDict,
            GenericUpdateColumnTypedDict,
        ]
        | GenericUpdateColumnTypedDict,
        *,
        where_condition: str,
        where_params: Sequence[Any],
    ):
        self._table = table
        self._columns = cast(
            dict,
            columns.model_dump(exclude_unset=True, by_alias=True)
            if isinstance(columns, BaseModel)
            else {
                (
                    cast(str, table.model_fields[k].alias)
                    if table.model_fields[k].alias is not None
                    else k
                ): v
                for k, v in columns.items()
            },
        )
        self._where_condition = where_condition
        self._where_params = where_params

    @override
    def build(self) -> QueryWithParams:
        values = ", ".join([f"{key} = %s" for key in self._columns.keys()])

        query = f"UPDATE {full_table_name(self._table)} SET {values} WHERE {self._where_condition}".strip()

        return QueryWithParams(
            query,
            tuple(itertools.chain(self._columns.values(), self._where_params)),
        )

    @override
    def execute(self, cursor: Cursor) -> None:
        query, params = self.build()

        execute_with_tag(
            UpdateTag[self._table],  # type: ignore
            cursor,
            query,
            params,
        )



from typing import Generic, TypeVar

T = TypeVar("T")
U = TypeVar("U")


class TypeMissMatch(Generic[T, U]):
    """
    Type Hint for Type Miss Match.
    """


class UseIsInsteadOfEq:
    """
    Please use `.is_` instead of `==`.
    """


class UseIsNotInsteadOfNe:
    """
    Please use `.is_.not_` instead of `!=`.
    """



from typing import Type, cast

from snowman.relation.table import (
    GenericColumnAccessor,
    GenericInsertColumnTypedDict,
    GenericOrderItemAccessor,
    GenericUpdateColumnTypedDict,
    Table,
)
from snowman.relation.table_like import GenericTableLike
from snowman.relation.view import View

from .builder.condition.condition import Condition
from .builder.condition.group_condition import GroupCondition
from .column import _InternalColumnAccessor
from .order_item import _InternalOrderItemAccessor


def group(condition: Condition) -> GroupCondition:
    return GroupCondition(condition)


def column(
    table: Type[
        Table[
            GenericTableLike,
            GenericColumnAccessor,
            GenericOrderItemAccessor,
            GenericInsertColumnTypedDict,
            GenericUpdateColumnTypedDict,
        ]
        | View[GenericColumnAccessor, GenericOrderItemAccessor]
    ],
) -> GenericColumnAccessor:
    """
    Extract column information from the table.

    #### Usage:
        ```python
        repr(column(User).id) == "database.schema.users.id"
        ```
    """
    return cast(GenericColumnAccessor, _InternalColumnAccessor(table))


def order(
    table: Type[
        Table[
            GenericTableLike,
            GenericColumnAccessor,
            GenericOrderItemAccessor,
            GenericInsertColumnTypedDict,
            GenericUpdateColumnTypedDict,
        ]
        | View[GenericColumnAccessor, GenericOrderItemAccessor]
    ],
) -> GenericOrderItemAccessor:
    """
    Extract column information from the table.

    #### Usage:
        ```python
        repr(column(User).id) == "database.schema.users.id"
        ```
    """
    return cast(GenericOrderItemAccessor, _InternalOrderItemAccessor(table))



from typing import Any, Callable, Generic, Sequence, Type, cast, overload

from typing_extensions import override

from snowman._cursor import Cursor
from snowman.query.builder.condition.condition import Condition
from snowman.query.builder.context.order_by_context import OrderByContext
from snowman.query.builder.context.where_context import WhereContext
from snowman.query.order_item import OrderItem
from snowman.relation import full_table_name
from snowman.relation.table import (
    GenericColumnAccessor,
    GenericInsertColumnTypedDict,
    GenericOrderItemAccessor,
    GenericUpdateColumnTypedDict,
    Table,
)
from snowman.relation.table_like import GenericTableLike
from snowman.relation.view import View

from ._builder import QueryBuilder, QueryWithParams, execute


class SelectQueryBuilder:
    def __init__(self):
        pass

    def from_(
        self,
        table: Type[
            Table[
                GenericTableLike,
                GenericColumnAccessor,
                GenericOrderItemAccessor,
                GenericInsertColumnTypedDict,
                GenericUpdateColumnTypedDict,
            ]
            | View[
                GenericColumnAccessor,
                GenericOrderItemAccessor,
            ]
        ],
        /,
    ) -> "SelectFromQueryBuilder[GenericTableLike, GenericColumnAccessor, GenericOrderItemAccessor, GenericInsertColumnTypedDict, GenericUpdateColumnTypedDict]":
        return SelectFromQueryBuilder(table)


class SelectFinalQueryBuilder(
    QueryBuilder[
        "SelectCursor[GenericTableLike, GenericColumnAccessor, GenericOrderItemAccessor, GenericInsertColumnTypedDict, GenericUpdateColumnTypedDict]"
    ],
    Generic[
        GenericTableLike,
        GenericColumnAccessor,
        GenericOrderItemAccessor,
        GenericInsertColumnTypedDict,
        GenericUpdateColumnTypedDict,
    ],
):
    def __init__(
        self,
        table: Type[
            Table[
                GenericTableLike,
                GenericColumnAccessor,
                GenericOrderItemAccessor,
                GenericInsertColumnTypedDict,
                GenericUpdateColumnTypedDict,
            ]
            | View[
                GenericColumnAccessor,
                GenericOrderItemAccessor,
            ]
        ],
        *,
        where_condition: str | None = None,
        where_params: Sequence[Any] | None = None,
        order_by_condition: str | None = None,
        order_by_params: Sequence[Any] | None = None,
        offset: int | None = None,
        limit: int | None = None,
    ):
        self._table = table
        self._where_condition = where_condition
        self._where_params = tuple(where_params) if where_params else ()
        self._order_by_condition = order_by_condition
        self._order_by_params = tuple(order_by_params) if order_by_params else ()
        self._offset = offset
        self._limit = limit

    @override
    def build(self) -> QueryWithParams:
        where_clause = (
            f" WHERE {self._where_condition}" if self._where_condition else ""
        )
        order_by_clause = (
            f" ORDER BY {self._order_by_condition}" if self._order_by_condition else ""
        )
        limit_clause = " LIMIT %s" if self._limit is not None else ""
        limit_params = (self._limit,) if self._limit is not None else ()
        offset_clause = " OFFSET %s" if self._offset is not None else ""
        offset_params = (self._offset,) if self._offset is not None else ()

        query = (
            f"SELECT * FROM {full_table_name(self._table)}"
            + where_clause
            + order_by_clause
            + limit_clause
            + offset_clause
        )

        return QueryWithParams(
            query,
            self._where_params + self._order_by_params + limit_params + offset_params,
        )

    @override
    def execute(
        self, cursor: Cursor
    ) -> "SelectCursor[GenericTableLike, GenericColumnAccessor, GenericOrderItemAccessor, GenericInsertColumnTypedDict, GenericUpdateColumnTypedDict]":
        query, params = self.build()

        return SelectCursor(execute(cursor, query, params), self._table)


class SelectFromWhereOrderByLimitQueryBuilder(
    SelectFinalQueryBuilder[
        GenericTableLike,
        GenericColumnAccessor,
        GenericOrderItemAccessor,
        GenericInsertColumnTypedDict,
        GenericUpdateColumnTypedDict,
    ]
):
    def __init__(
        self,
        table: Type[
            Table[
                GenericTableLike,
                GenericColumnAccessor,
                GenericOrderItemAccessor,
                GenericInsertColumnTypedDict,
                GenericUpdateColumnTypedDict,
            ]
            | View[
                GenericColumnAccessor,
                GenericOrderItemAccessor,
            ]
        ],
        where_condition: str | None = None,
        where_params: Sequence[Any] | None = None,
        order_by_condition: str | None = None,
        order_by_params: Sequence[Any] | None = None,
        limit: int | None = None,
    ):
        super().__init__(
            table,
            where_condition=where_condition,
            where_params=where_params,
            order_by_condition=order_by_condition,
            order_by_params=order_by_params,
            limit=limit,
        )

    def offset(
        self, offset: int | None
    ) -> "SelectFinalQueryBuilder[GenericTableLike, GenericColumnAccessor, GenericOrderItemAccessor, GenericInsertColumnTypedDict, GenericUpdateColumnTypedDict]":
        return SelectFinalQueryBuilder(
            self._table,
            where_condition=self._where_condition,
            where_params=self._where_params,
            order_by_condition=self._order_by_condition,
            order_by_params=self._order_by_params,
            limit=self._limit,
            offset=offset,
        )


class SelectFromWhereOrderByQueryBuilder(
    SelectFinalQueryBuilder[
        GenericTableLike,
        GenericColumnAccessor,
        GenericOrderItemAccessor,
        GenericInsertColumnTypedDict,
        GenericUpdateColumnTypedDict,
    ]
):
    def __init__(
        self,
        table: Type[
            Table[
                GenericTableLike,
                GenericColumnAccessor,
                GenericOrderItemAccessor,
                GenericInsertColumnTypedDict,
                GenericUpdateColumnTypedDict,
            ]
            | View[
                GenericColumnAccessor,
                GenericOrderItemAccessor,
            ]
        ],
        where_condition: str | None = None,
        where_params: Sequence[Any] | None = None,
        order_by_condition: str | None = None,
        order_by_params: Sequence[Any] | None = None,
    ):
        super().__init__(
            table,
            where_condition=where_condition,
            where_params=where_params,
            order_by_condition=order_by_condition,
            order_by_params=order_by_params,
        )

    def limit(
        self, limit: int | None
    ) -> "SelectFromWhereOrderByLimitQueryBuilder[GenericTableLike, GenericColumnAccessor,  GenericOrderItemAccessor, GenericInsertColumnTypedDict, GenericUpdateColumnTypedDict]":
        return SelectFromWhereOrderByLimitQueryBuilder(
            self._table,
            where_condition=self._where_condition,
            where_params=self._where_params,
            order_by_condition=self._order_by_condition,
            order_by_params=self._order_by_params,
            limit=limit,
        )


class SelectFromWhereOrderQueryBuilder(
    Generic[
        GenericTableLike,
        GenericColumnAccessor,
        GenericOrderItemAccessor,
        GenericInsertColumnTypedDict,
        GenericUpdateColumnTypedDict,
    ]
):
    def __init__(
        self,
        table: Type[
            Table[
                GenericTableLike,
                GenericColumnAccessor,
                GenericOrderItemAccessor,
                GenericInsertColumnTypedDict,
                GenericUpdateColumnTypedDict,
            ]
            | View[
                GenericColumnAccessor,
                GenericOrderItemAccessor,
            ]
        ],
        where_condition: str | None = None,
        where_params: Sequence[Any] | None = None,
    ):
        self._table = table
        self._where_condition = where_condition
        self._where_params = tuple(where_params) if where_params else ()

    @overload
    def by(
        self,
        condition: Callable[
            [
                OrderByContext[
                    GenericTableLike,
                    GenericColumnAccessor,
                    GenericOrderItemAccessor,
                    GenericInsertColumnTypedDict,
                    GenericUpdateColumnTypedDict,
                ]
            ],
            Sequence[OrderItem[GenericTableLike, Any, Any]]
            | OrderItem[GenericTableLike, Any, Any],
        ],
        /,
    ) -> "SelectFromWhereOrderByQueryBuilder[GenericTableLike, GenericColumnAccessor, GenericOrderItemAccessor, GenericInsertColumnTypedDict, GenericUpdateColumnTypedDict]": ...

    @overload
    def by(
        self,
        condition: Sequence[OrderItem[GenericTableLike, Any, Any]]
        | OrderItem[GenericTableLike, Any, Any],
        /,
    ) -> "SelectFromWhereOrderByQueryBuilder[GenericTableLike, GenericColumnAccessor, GenericOrderItemAccessor, GenericInsertColumnTypedDict, GenericUpdateColumnTypedDict]": ...

    @overload
    def by(
        self,
        condition: str,
        params: Sequence[Any] | None = None,
        /,
    ) -> "SelectFromWhereOrderByQueryBuilder[GenericTableLike, GenericColumnAccessor, GenericOrderItemAccessor, GenericInsertColumnTypedDict, GenericUpdateColumnTypedDict]": ...

    def by(
        self,
        condition: Callable[
            [
                OrderByContext[
                    GenericTableLike,
                    GenericColumnAccessor,
                    GenericOrderItemAccessor,
                    GenericInsertColumnTypedDict,
                    GenericUpdateColumnTypedDict,
                ]
            ],
            Sequence[OrderItem[GenericTableLike, Any, Any]]
            | OrderItem[GenericTableLike, Any, Any],
        ]
        | Sequence[OrderItem[GenericTableLike, Any, Any]]
        | OrderItem[GenericTableLike, Any, Any]
        | str,
        params: Sequence[Any] | None = None,
        /,
    ) -> "SelectFromWhereOrderByQueryBuilder[GenericTableLike, GenericColumnAccessor, GenericOrderItemAccessor, GenericInsertColumnTypedDict, GenericUpdateColumnTypedDict]":
        if callable(condition):
            condition = condition(OrderByContext(self._table))

        if not isinstance(condition, str):
            if isinstance(condition, OrderItem):
                condition, params = condition.to_sql()
            else:
                items: list[str] = []
                params = []
                for key in condition:
                    sql = key.to_sql()
                    items.append(sql[0])
                    params.extend(sql[1])

                condition = ", ".join(items)

        return SelectFromWhereOrderByQueryBuilder(
            self._table,
            where_condition=self._where_condition,
            where_params=self._where_params,
            order_by_condition=condition,
            order_by_params=params or (),
        )


class SelectFromWhereQueryBuilder(
    SelectFromWhereOrderByQueryBuilder[
        GenericTableLike,
        GenericColumnAccessor,
        GenericOrderItemAccessor,
        GenericInsertColumnTypedDict,
        GenericUpdateColumnTypedDict,
    ],
):
    def __init__(
        self,
        table: Type[
            Table[
                GenericTableLike,
                GenericColumnAccessor,
                GenericOrderItemAccessor,
                GenericInsertColumnTypedDict,
                GenericUpdateColumnTypedDict,
            ]
            | View[
                GenericColumnAccessor,
                GenericOrderItemAccessor,
            ]
        ],
        where_condition: str | None = None,
        where_params: Sequence[Any] | None = None,
    ):
        super().__init__(
            table,
            where_condition=where_condition,
            where_params=where_params,
        )

    @property
    def order(
        self,
    ) -> "SelectFromWhereOrderQueryBuilder[GenericTableLike, GenericColumnAccessor, GenericOrderItemAccessor, GenericInsertColumnTypedDict, GenericUpdateColumnTypedDict]":
        return SelectFromWhereOrderQueryBuilder(
            self._table,
            where_condition=self._where_condition,
            where_params=self._where_params,
        )


class SelectFromQueryBuilder(
    SelectFromWhereQueryBuilder[
        GenericTableLike,
        GenericColumnAccessor,
        GenericOrderItemAccessor,
        GenericInsertColumnTypedDict,
        GenericUpdateColumnTypedDict,
    ]
):
    def __init__(
        self,
        table: Type[
            Table[
                GenericTableLike,
                GenericColumnAccessor,
                GenericOrderItemAccessor,
                GenericInsertColumnTypedDict,
                GenericUpdateColumnTypedDict,
            ]
            | View[
                GenericColumnAccessor,
                GenericOrderItemAccessor,
            ]
        ],
    ):
        super().__init__(table)

    @overload
    def where(
        self,
        condition: Callable[[WhereContext[GenericColumnAccessor]], Condition],
        /,
    ) -> "SelectFromWhereQueryBuilder[GenericTableLike, GenericColumnAccessor, GenericOrderItemAccessor, GenericInsertColumnTypedDict, GenericUpdateColumnTypedDict]": ...

    @overload
    def where(
        self,
        condition: Condition,
        /,
    ) -> "SelectFromWhereQueryBuilder[GenericTableLike, GenericColumnAccessor, GenericOrderItemAccessor, GenericInsertColumnTypedDict, GenericUpdateColumnTypedDict]": ...

    @overload
    def where(
        self,
        condition: str,
        params: Sequence[Any] | None = None,
        /,
    ) -> "SelectFromWhereQueryBuilder[GenericTableLike, GenericColumnAccessor, GenericOrderItemAccessor, GenericInsertColumnTypedDict, GenericUpdateColumnTypedDict]": ...

    def where(
        self,
        condition: Callable[[WhereContext[GenericColumnAccessor]], Condition]
        | Condition
        | str,
        params: Sequence[Any] | None = None,
        /,
    ) -> "SelectFromWhereQueryBuilder[GenericTableLike, GenericColumnAccessor, GenericOrderItemAccessor, GenericInsertColumnTypedDict, GenericUpdateColumnTypedDict]":
        if callable(condition):
            condition, params = condition(WhereContext(self._table)).to_sql()
        elif isinstance(condition, Condition):
            condition, params = condition.to_sql()

        return SelectFromWhereQueryBuilder(
            self._table,
            where_condition=condition,
            where_params=params or (),
        )


class SelectCursor(
    Generic[
        GenericTableLike,
        GenericColumnAccessor,
        GenericOrderItemAccessor,
        GenericInsertColumnTypedDict,
        GenericUpdateColumnTypedDict,
    ]
):
    def __init__(
        self,
        cursor: Cursor,
        table: Type[
            Table[
                GenericTableLike,
                GenericColumnAccessor,
                GenericOrderItemAccessor,
                GenericInsertColumnTypedDict,
                GenericUpdateColumnTypedDict,
            ]
            | View[
                GenericColumnAccessor,
                GenericOrderItemAccessor,
            ]
        ],
    ):
        self._cursor = cursor
        self._table = table

    @property
    def rowcount(self) -> int | None:
        return self._cursor.rowcount

    def fetchall(self) -> list[GenericTableLike]:
        return [
            cast(GenericTableLike, _map_row(self._table, row))
            for row in self._cursor.fetchall()
        ]

    def fetchmany(self, size: int | None = None) -> list[GenericTableLike]:
        return [
            cast(GenericTableLike, _map_row(self._table, row))
            for row in self._cursor.fetchmany(size)
        ]

    def fetchone(self) -> GenericTableLike | None:
        row = self._cursor.fetchone()

        if row is None or len(row) == 0:
            return None

        return cast(
            GenericTableLike,
            _map_row(self._table, row),
        )

    def __iter__(self):
        return self

    def __next__(self) -> GenericTableLike:
        row = self._cursor.fetchone()

        if row is None:
            raise StopIteration

        return cast(
            GenericTableLike,
            _map_row(self._table, row),
        )


def _map_row(row_type: Type[GenericTableLike], row: Any) -> GenericTableLike:
    return cast(
        GenericTableLike,
        row_type.model_validate(
            {key: data for key, data in zip(row_type.model_fields.keys(), row)}
        ),
    )



from typing import Any, Callable, Generic, Sequence, Type, overload

from typing_extensions import override

from snowman._cursor import Cursor
from snowman._features import DeleteTag
from snowman.query.builder.condition.condition import Condition
from snowman.query.builder.context.where_context import WhereContext
from snowman.relation import full_table_name
from snowman.relation.table import (
    GenericColumnAccessor,
    GenericInsertColumnTypedDict,
    GenericOrderItemAccessor,
    GenericTable,
    GenericUpdateColumnTypedDict,
    Table,
)

from ._builder import QueryBuilder, QueryWithParams, execute_with_tag


class DeleteQueryBuilder:
    def from_(
        self,
        table: Type[
            Table[
                GenericTable,
                GenericColumnAccessor,
                GenericOrderItemAccessor,
                GenericInsertColumnTypedDict,
                GenericUpdateColumnTypedDict,
            ]
        ],
        /,
    ) -> "DeleteFromStatement[GenericColumnAccessor]":
        return DeleteFromStatement(table)


class DeleteFromStatement(Generic[GenericColumnAccessor]):
    def __init__(
        self,
        table: Type[
            Table[
                GenericTable,
                GenericColumnAccessor,
                GenericOrderItemAccessor,
                GenericInsertColumnTypedDict,
                GenericUpdateColumnTypedDict,
            ]
        ],
    ):
        self._table = table

    @overload
    def where(
        self,
        condition: Callable[[WhereContext[GenericColumnAccessor]], Condition],
        /,
    ) -> "DeleteFromWhereQueryBuilder": ...

    @overload
    def where(
        self,
        condition: Condition,
        /,
    ) -> "DeleteFromWhereQueryBuilder": ...

    @overload
    def where(
        self,
        condition: str,
        params: Sequence[Any] | None = None,
        /,
    ) -> "DeleteFromWhereQueryBuilder": ...

    def where(
        self,
        condition: Callable[[WhereContext[GenericColumnAccessor]], Condition]
        | Condition
        | str,
        params: Sequence[Any] | None = None,
        /,
    ) -> "DeleteFromWhereQueryBuilder":
        """
        Specify the condition of the where clause.

        Query parameters only support positional placeholders, so specify them with `%s`.

        e.g)
            - `.where(lambda c: (c(User).id == 1).and_(c(User).name == "Alice"))`
            - `.where((column(User).id == 1).and_(column(User).name == "Alice"))`
            - `.where("id = %s AND name = %s", [1, "Alice"])`
        """
        if callable(condition):
            condition, params = condition(WhereContext(self._table)).to_sql()

        elif isinstance(condition, Condition):
            condition, params = condition.to_sql()

        return DeleteFromWhereQueryBuilder(self._table, condition, params or ())


class DeleteFromWhereQueryBuilder(QueryBuilder[None]):
    def __init__(
        self,
        table: Type[
            Table[
                GenericTable,
                GenericColumnAccessor,
                GenericOrderItemAccessor,
                GenericInsertColumnTypedDict,
                GenericUpdateColumnTypedDict,
            ]
        ],
        where_condition: str,
        where_params: Sequence[Any],
    ):
        self._table = table
        self._where_condition = where_condition
        self._where_params = where_params

    @override
    def build(self) -> QueryWithParams:
        query = f"DELETE FROM {full_table_name(self._table)} WHERE {self._where_condition}".strip()

        return QueryWithParams(query, tuple(self._where_params))

    @override
    def execute(self, cursor: Cursor) -> None:
        query, params = self.build()

        execute_with_tag(
            DeleteTag[self._table],  # type: ignore
            cursor,
            query,
            params,
        )



from abc import ABC, abstractmethod
from typing import Any, Generic, NamedTuple, TypeVar

from snowman._cursor import Cursor
from snowman._features import USE_TURU, TuruSnowflakeCursor

GenericExecuteResult = TypeVar("GenericExecuteResult")


class QueryWithParams(NamedTuple):
    query: str
    params: tuple[Any, ...]


class QueryBuilder(Generic[GenericExecuteResult], ABC):
    @abstractmethod
    def build(self) -> QueryWithParams: ...

    @abstractmethod
    def execute(self, cursor: Cursor, /) -> GenericExecuteResult: ...


def execute(cursor: Cursor, query: str, params: tuple[Any, ...]):
    cursor.execute(query, params)

    return cursor


def execute_with_tag(
    tag,
    cursor: Cursor,
    query: str,
    params: tuple[Any, ...],
):
    if USE_TURU and isinstance(cursor, TuruSnowflakeCursor):
        cursor.execute_with_tag(tag, query, params)

    else:
        cursor.execute(query, params)


def executemany_with_tag(
    tag,
    cursor: Cursor,
    query: str,
    params: tuple[Any, ...],
):
    if USE_TURU and isinstance(cursor, TuruSnowflakeCursor):
        cursor.executemany_with_tag(tag, query, params)

    else:
        cursor.executemany(query, params)



from abc import ABC, abstractmethod
from typing import Any, NamedTuple


class OperationWithParams(NamedTuple):
    operation: str
    params: tuple[Any, ...]


class ToSql(ABC):
    @abstractmethod
    def to_sql(self) -> OperationWithParams: ...



from typing import Generic, Type

from snowman.query.builder.condition.condition import Condition
from snowman.query.builder.condition.group_condition import GroupCondition
from snowman.query.expression import column, group
from snowman.relation.table import (
    GenericColumnAccessor,
    GenericInsertColumnTypedDict,
    GenericOrderItemAccessor,
    GenericUpdateColumnTypedDict,
    Table,
)
from snowman.relation.table_like import GenericTableLike
from snowman.relation.view import View


class WhereContext(Generic[GenericColumnAccessor]):
    def __init__(
        self,
        table: Type[
            Table[
                GenericTableLike,
                GenericColumnAccessor,
                GenericOrderItemAccessor,
                GenericInsertColumnTypedDict,
                GenericUpdateColumnTypedDict,
            ]
            | View[
                GenericColumnAccessor,
                GenericOrderItemAccessor,
            ]
        ],
    ) -> None:
        self._table = table

    def __call__(
        self,
        table: Type[
            Table[
                GenericTableLike,
                GenericColumnAccessor,
                GenericOrderItemAccessor,
                GenericInsertColumnTypedDict,
                GenericUpdateColumnTypedDict,
            ]
        ],
    ) -> GenericColumnAccessor:
        return column(table)

    def group(self, condition: Condition, /) -> GroupCondition:
        return group(condition)

    @property
    def self(self) -> GenericColumnAccessor:
        return column(self._table)



from typing import TYPE_CHECKING

from snowman._generic import PyType
from snowman.query.builder.condition.condition import Condition
from snowman.query.to_sql import OperationWithParams

if TYPE_CHECKING:
    from snowman.query.column import Column, GenericColumnName
    from snowman.relation.table_like import GenericTableLike


class LeCondition(Condition):
    def __init__(
        self, base: "Column[GenericTableLike, GenericColumnName, PyType]", value: PyType
    ):
        self._base = base
        self._value = value

    def to_sql(self) -> OperationWithParams:
        return OperationWithParams(
            operation=f"{self._base} <= %s", params=(self._value,)
        )



from typing import Generic, Sequence, Type

from typing_extensions import override

import snowman._features
from snowman._cursor import Cursor, _get_snowflake_connection
from snowman.exception import snowmanNotDataFrameAvailableError
from snowman.relation import full_table_name, table_column_names, table_columns_dict
from snowman.relation.table import (
    GenericColumnAccessor,
    GenericInsertColumnTypedDict,
    GenericOrderItemAccessor,
    GenericTable,
    GenericUpdateColumnTypedDict,
    Table,
)

from ._builder import (
    QueryBuilder,
    QueryWithParams,
    execute_with_tag,
    executemany_with_tag,
)


class InsertQueryBuilder:
    """
    Insert records into a table.

    >>> from typing import TypedDict, Literal
    >>> import pydantic
    >>> from snowman.query import insert
    >>> from snowman import Column, ColumnOrderItem
    >>>
    >>> class _UserColumnAccessor(TypedDict):
    ...     id: Column["User", Literal["id"], int]
    ...     name: Column["User", Literal["age"], str]
    ...
    >>> class _UserOrderItemAccessor(TypedDict):
    ...     id: ColumnOrderItem["User", Literal["id"], int]
    ...     name: ColumnOrderItem["User", Literal["age"], str]
    ...
    >>> class _UserInsertColumns(TypedDict):
    ...     id: int
    ...     name: str
    ...
    >>> class _UserUpdateColumns(TypedDict, total=False):
    ...     id: int
    ...     name: str
    ...
    >>> @snowman.table("database", "schema", "users")
    ... class User(
    ...     snowman.Table["User", "_UserColumnAccessor", "_UserOrderItemAccessor", "_UserInsertColumns", "_UserUpdateColumns"]
    ... ):
    ...     id: int
    ...     name: str
    ...
    >>> query, _ = (
    ...     insert.into(
    ...         User
    ...     ).values(
    ...         {"id": 1, "name": "Alice"}
    ...     )
    ... ).build()
    >>>
    >>> print(query)
    INSERT INTO database.schema.users ( id, name ) VALUES ( %s, %s )
    """

    @property
    def overwrite(self) -> "InsertOverwriteQueryBuilder":
        return InsertOverwriteQueryBuilder()

    def into(
        self,
        table: Type[
            Table[
                GenericTable,
                GenericColumnAccessor,
                GenericOrderItemAccessor,
                GenericInsertColumnTypedDict,
                GenericUpdateColumnTypedDict,
            ]
        ],
        /,
    ):
        return InsertIntoQueryBuilder(
            table,
            _columns_type=table.__insert_columns__,  # type: ignore
        )


class InsertOverwriteQueryBuilder:
    def into(
        self,
        table: Type[
            Table[
                GenericTable,
                GenericColumnAccessor,
                GenericOrderItemAccessor,
                GenericInsertColumnTypedDict,
                GenericUpdateColumnTypedDict,
            ]
        ],
        /,
    ):
        return InsertIntoQueryBuilder(
            table,
            overwtire=True,
            _columns_type=table.__insert_columns__,  # type: ignore
        )


class InsertIntoQueryBuilder(Generic[GenericTable, GenericInsertColumnTypedDict]):
    def __init__(
        self,
        table: Type[GenericTable],
        *,
        overwtire: bool = False,
        _columns_type: Type[GenericInsertColumnTypedDict] | None,
    ) -> None:
        self._table = table
        self._overwrite = overwtire

    def values(
        self,
        values: GenericTable
        | GenericInsertColumnTypedDict
        | Sequence[GenericTable | GenericInsertColumnTypedDict]
        | snowman._features.PandasDataFrame,
        /,
    ) -> "InsertIntoValuesQueryBuilder[GenericTable, GenericInsertColumnTypedDict]":
        if isinstance(values, snowman._features.PandasDataFrame):
            dataframe = values
            values = ()

        else:
            dataframe = None
            if not isinstance(values, Sequence):
                values = (values,)

        return InsertIntoValuesQueryBuilder(
            self._table,
            values,
            dataframe=dataframe,
            overwrite=self._overwrite,
        )


class InsertIntoValuesQueryBuilder(
    QueryBuilder[None], Generic[GenericTable, GenericInsertColumnTypedDict]
):
    def __init__(
        self,
        table: type[GenericTable],
        values: Sequence[GenericTable | GenericInsertColumnTypedDict],
        /,
        *,
        dataframe: snowman._features.PandasDataFrame | None = None,
        overwrite: bool = False,
    ):
        self._table = table
        self._values = values
        self._dataframe = dataframe
        self._overwrite = overwrite
        self._use_execute_many = len(values) != 1

    @override
    def build(self) -> QueryWithParams:
        if self._dataframe is not None:
            raise snowmanNotDataFrameAvailableError()

        overwrite = "OVERWRITE " if self._overwrite else ""
        names = table_column_names(self._table)
        keys = ", ".join(names)
        values = ", ".join("%s" for _ in names)

        query = f"INSERT {overwrite}INTO {full_table_name(self._table)} ( {keys} ) VALUES ( {values} )"

        return QueryWithParams(
            query,
            (
                tuple(
                    tuple(table_columns_dict(value).values()) for value in self._values
                )
                if self._use_execute_many
                else tuple(table_columns_dict(self._values[0]).values())
            ),
        )

    @override
    def execute(self, cursor: Cursor, /) -> None:
        if self._dataframe is not None:
            from snowflake.connector.pandas_tools import write_pandas

            write_pandas(
                conn=_get_snowflake_connection(cursor),
                df=self._dataframe,
                table_name=self._table.__table_name__,
                database=self._table.__database_name__,
                schema=self._table.__schema_name__,
            )

        elif self._use_execute_many:
            query, params = self.build()
            executemany_with_tag(
                snowman._features.InsertTag[self._table],  # type: ignore
                cursor,
                query,
                params,
            )

        else:
            query, params = self.build()
            execute_with_tag(
                snowman._features.InsertTag[self._table],  # type: ignore
                cursor,
                query,
                params,
            )
